import numpy as np
import os
os.environ['TRANSFORMERS_CACHE'] = 'cache'  # Must be before transformers import.
import pandas as pd
import sklearn
from sklearn import cluster, linear_model, metrics, multiclass
from sklearn.model_selection import train_test_split
import torch
import tqdm
import transformers

from . import utils
from .task import Task


class ClassificationTask(Task):
    def __init__(self, train, val, test):
        print(f'\n(1/3) PROCESSING DATA...\n')
        train = train if train is not None else pd.DataFrame([], columns=['text', 'label'])
        val = val if val is not None else pd.DataFrame([], columns=['text', 'label'])
        (
            self.train_embeds,
            self.train_labels,
            self.val_embeds,
            self.val_labels,
            self.test_embeds
        ) = self._load_data(train, val, test)

        self.classes = list(set(self.train_labels + self.val_labels))
        self.train_labels = sklearn.preprocessing.label_binarize(self.train_labels, classes=self.classes).tolist()
        if len(self.train_labels[0]) == 1:  # Quirk of label_binarize: binary cases must be expanded from a single column vector.
            self.train_labels = [[1, 0] if x == [0] else [0, 1] for x in self.train_labels]
        if len(self.val_labels) > 0:
            self.val_labels = sklearn.preprocessing.label_binarize(self.val_labels, classes=self.classes).tolist()
            if len(self.val_labels[0]) == 1:  # Quirk of label_binarize: binary cases must be expanded from a single column vector.
                self.val_labels = [[1, 0] if x == [0] else [0, 1] for x in self.val_labels]

        self.model = self._load_model()
    
    def _load_data(self, train, val, test):
        text = train.text.tolist() + val.text.tolist() + test.text.tolist()  # Must embed togeter to match sequence length.

        if len(max(text, key=len).split()) > 400:
            print('''
***

WARNING
At least one of your texts is long so it may have been truncated.
In general, this is okay. If you wish to use all your data, we
suggest you split your long texts into multiple lines. Try to remain
under 400 words per text.

***
''')
        print('  - Data Processing Step 1 of 2...')
        # Tokenize.
        tokenizer = transformers.BertTokenizer.from_pretrained('bert-base-uncased')

        tokens = []
        for t in tqdm.tqdm(text):
            token_ids = tokenizer(t, return_tensors='pt', padding=True, truncation=True)
            tokens.append(token_ids)

        # Embed.
        print('  - Data Processing Step 2 of 2...')
        embeds = []
        embed_model = transformers.BertModel.from_pretrained('bert-base-uncased')
        embed_model.eval()
        with torch.no_grad():
            for token in tqdm.tqdm(tokens):
                embed = embed_model(**token)[0]
                embeds.append(embed.mean(1).squeeze())
        embeds = torch.stack(embeds, 0)

        return embeds[:len(train)], train.label.tolist(), embeds[len(train):len(train) + len(val)], val.label.tolist(), embeds[len(train) + len(val):]

    def _load_model(self):
        return multiclass.OneVsRestClassifier(linear_model.LogisticRegression(random_state=0, max_iter=1000))
    
    def train(self):
        print(f'\n(2/3) LEARNING...\n')
        self.model.fit(self.train_embeds, self.train_labels)

    def evaluate(self):
        if len(self.val_embeds) == 0:
            return None

        val_preds = self.model.predict(self.val_embeds)
        val_probs = self.model.predict_proba(self.val_embeds)

        accs = {}
        precs = {}
        recs = {}
        f1s = {}
        aurocs = {}
        auprs = {}

        # Get one-vs-all classification metrics for each class.
        print('\nGETTING METRICS...\n')
        for i, class_name in tqdm.tqdm(enumerate(self.classes)):
            class_bin = [0 for _ in range(len(self.classes))]
            class_bin[i] = 1

            labels_bin = [1 if j == class_bin else 0 for j in self.val_labels]
            preds_bin = [1 if j == class_bin else 0 for j in val_preds.tolist()]

            accs[class_name] = metrics.accuracy_score(labels_bin, preds_bin).item()
            precs[class_name] = metrics.precision_score(labels_bin, preds_bin, zero_division=0).item()
            recs[class_name] = metrics.recall_score(labels_bin, preds_bin, zero_division=0).item()
            f1s[class_name] = metrics.f1_score(labels_bin, preds_bin, zero_division=0).item()

            fpr, tpr, _ = metrics.roc_curve(labels_bin, val_probs[:,i])
            aurocs[class_name] = metrics.auc(fpr, tpr).item()

            precision, recall, _ = metrics.precision_recall_curve(labels_bin, val_probs[:,i])
            auprs[class_name] = metrics.auc(recall, precision).item()

        return {
            'acc': accs,
            'prec': precs,
            'rec': recs,
            'f1': f1s,
            'auroc': aurocs,
            'aupr': auprs,
        }
    
    def predict(self):
        print(f'\n(3/3) CREATING TAILWIZ LABELS...\n')
        out_predictions = []
        for i in tqdm.tqdm(range(self.test_embeds.shape[0])):
            prediction = self.model.predict(self.test_embeds[i].unsqueeze(0))
            class_i = np.argmax(prediction[0])
            out_predictions.append(self.classes[class_i])
        return out_predictions


class KMeansClassificationTask(ClassificationTask):
    def __init__(self, test):
        print(f'\n(1/2) PROCESSING DATA...\n')
        (_, _, _, _, self.test_embeds) = self._load_data(
            pd.DataFrame([], columns=['text', 'label']),
            pd.DataFrame([], columns=['text', 'label']),
            test
        )
        self.model = self._load_model()
    
    def _load_data(self, train, val, test):
        return super()._load_data(train, val, test)

    def _load_model(self):
        return cluster.KMeans(n_clusters=2, random_state=0, max_iter=1000)
    
    def train(self):
        self.model.fit(self.test_embeds)
    
    def evaluate(self):
        # KMeans task is default task when no train/val data is provided.
        # Thus, it should not have an evaluate function.
        pass
    
    def predict(self):
        print(f'\n(2/2) CREATING TAILWIZ LABELS...\n')
        out_predictions = []
        for i in tqdm.tqdm(range(self.test_embeds.shape[0])):
            prediction = self.model.predict(self.test_embeds[i].unsqueeze(0))
            out_predictions.append(prediction[0])
        return out_predictions  # self.model.predict(self.test_embeds)


def classify(to_classify, labeled_examples=None, output_metrics=False):
    assert isinstance(to_classify, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
    assert 'text' in to_classify.columns, 'Make sure the text column in your pandas DataFrame is named "text".'

    if labeled_examples is not None:
        assert isinstance(labeled_examples, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
        assert 'text' in labeled_examples.columns and 'label' in labeled_examples.columns, \
            'Make sure the text column in your pandas DataFrame is named "text" and the label column is named "label"'

    if output_metrics:
        assert labeled_examples is not None, 'In order to output an estimate of performance with output_metrics, labeled_examples must be provided.'

    # Perform KMeans if no training data is given.
    if labeled_examples is None:
        task = KMeansClassificationTask(to_classify)
        task.train()
        pred_results = task.predict()
        results = to_classify.copy()
        results['tailwiz_label'] = pred_results
        return results

    if len(labeled_examples) < 3:
        raise ValueError('labeled_examples has too few examples. At least 3 are required.')

    num_unique_classes = len(labeled_examples.label.unique())
    if num_unique_classes <= 1:
        raise ValueError('labeled_examples contains examples from just one class. Examples from at least 2 classes are required.')

    # Try 10 times to get a proper split. Sometimes, a split will cause all training examples to be
    # in the same class, which will error.
    classify_task_out = None
    split_attempt = 0
    while split_attempt < 10 and classify_task_out is None:
        train, val = sklearn.model_selection.train_test_split(labeled_examples, test_size=0.2)
        num_unique_classes_in_train = len(train.label.unique())
        if num_unique_classes_in_train < 2:
            split_attempt += 1
            continue
        classify_task_out = ClassificationTask(train, val, to_classify)
        classify_task_out.train()
    
    if classify_task_out is None:
        raise ValueError('''The provided labeled_examples examples were not diverse enough to estimate performance.
        Try balancing your labeled_examples examples by adding more examples of each class.''')
    
    pred_results = classify_task_out.predict()
    results = to_classify.copy()
    results['tailwiz_label'] = pred_results

    if output_metrics:
        metrics_out = classify_task_out.evaluate()

    print('\nDONE')
    return (results, metrics_out) if output_metrics else results



import os
os.environ['TRANSFORMERS_CACHE'] = 'cache'  # Must be before transformers import.
import pandas as pd
import torch
import tqdm
import transformers
import evaluate
from sklearn.model_selection import train_test_split

from .task import Task

class ParsingTask(Task):
    def __init__(self, train, val, test):
        self.num_steps = 2 if train is None else 3
        self.tokenizer = transformers.BertTokenizer.from_pretrained('bert-large-uncased-whole-word-masking-finetuned-squad')
        (
            self.train_dataset,
            self.val_dataset,
            self.test_dataset
        ) = self._load_data(train, val, test)
        self.model = self._load_model()

    def _load_data(self, train, val, test):
        print(f'\n(1/{self.num_steps}) PROCESSING DATA...\n')

        text = test.prompt.tolist() + test.context.tolist()
        if train is not None:
            text += train.prompt.tolist() + train.context.tolist() + train.label.tolist()
        if val is not None:
            text += val.prompt.tolist() + val.context.tolist() + val.label.tolist()
        if len(max(text, key=len).split()) > 400:
            print('''
***

WARNING
At least one of your texts is long so it may have been truncated.
In general, this is okay. If you wish to use all your data, we
suggest you split your long texts into multiple lines. Try to remain
under 400 words per text.

***
''')
        # Tokenize.
        with tqdm.tqdm(total=5) as pbar:
            train_tokens = None if train is None else self.tokenizer(train.prompt.tolist(), train.context.tolist(), train.label.tolist(), padding=True, truncation=True, return_tensors='pt')
            pbar.update()
            
            val_tokens = None if val is None else self.tokenizer(val.prompt.tolist(), val.context.tolist(), val.label.tolist(), padding=True, truncation=True, return_tensors='pt')
            pbar.update()

            test_tokens = self.tokenizer(test.prompt.tolist(), test.context.tolist(), padding=True, truncation=True, return_tensors='pt')
            pbar.update()

            # Helper function that finds the start and end positions of the label within the context.
            # By default sets the start and end positions to the first occurrence of the label.
            def get_target_start_end_pos(tokens, contexts, labels):
                target_start_pos = []
                target_end_pos = []
                for i, (text_tokenized, label_tokenized, context, label) in enumerate(zip(tokens['input_ids'], tokens['labels'], contexts, labels)):
                    if label_tokenized[0] == 101 and label_tokenized[1] == 102:
                        target_start_pos.append(0)
                        target_end_pos.append(0)
                    else:
                        label_tokenized_nopad = label_tokenized[label_tokenized.nonzero()].squeeze()
                        label_tokenized_nospectoks = label_tokenized_nopad[1:-1]

                        for j in range(len(text_tokenized) - len(label_tokenized_nospectoks)):  # Search from start for label
                            if (text_tokenized[j:j+len(label_tokenized_nospectoks)] == label_tokenized_nospectoks).all():
                                target_start_pos.append(j)
                                target_end_pos.append(j+len(label_tokenized_nospectoks))
                                break
                    if len(target_start_pos) != i + 1:
                        raise ValueError(f'''We could not find the label:
                            
                        {label}

in the context:
                        
                        {context}

because either the context was truncated due to its length and the
label appears later in the context, or because the label is not found
at all in the context.

To resolve, either: (A) only include labels where the label is extracted
exactly from the context as whole words, or (B) use the generate() function,
which does not require the labels do be found in the context.''')
                target_start_pos = torch.tensor(target_start_pos)
                target_end_pos = torch.tensor(target_end_pos)
                return target_start_pos, target_end_pos

            if train_tokens is not None:
                train_target_start_pos, train_target_end_pos = get_target_start_end_pos(train_tokens, train.context.tolist(), train.label.tolist())
                train_tokens['start_positions'] = train_target_start_pos
                train_tokens['end_positions'] =  train_target_end_pos
                train_tokens = {k:v for k, v in train_tokens.items() if k != 'labels'}
            pbar.update()

            if val_tokens is not None:
                val_target_start_pos, val_target_end_pos = get_target_start_end_pos(val_tokens, val.context.tolist(), val.label.tolist())
                val_tokens['start_positions'] = val_target_start_pos
                val_tokens['end_positions'] =  val_target_end_pos
                val_tokens = {k:v for k, v in val_tokens.items() if k != 'labels'}
            pbar.update()

        return train_tokens, val_tokens, test_tokens

    def _load_model(self):
        return transformers.BertForQuestionAnswering.from_pretrained('bert-large-uncased-whole-word-masking-finetuned-squad')

    def _decode_pos2strs(self, start_pos, end_pos, input_ids):
        strs = []
        for i, (j, k) in enumerate(zip(start_pos, end_pos)):
            pred_tokens = input_ids[i, j:k+1]
            strs.append(self.tokenizer.decode(pred_tokens, skip_special_tokens=True))
        return strs

    def train(self):
        print(f'\n(2/{self.num_steps}) LEARNING...\n')
        class BLUWWMFSDataset(torch.utils.data.Dataset):
            def __init__(self, data):
                self.data = data
            
            def __len__(self):
                return len(self.data['input_ids'])
            
            def __getitem__(self, index):
                return {k: self.data[k][index] for k in self.data.keys()}

        args = transformers.TrainingArguments(
            'cache/bert-qa',
            num_train_epochs=10,
            evaluation_strategy='epoch',
            save_strategy='epoch',
            load_best_model_at_end=True,
            metric_for_best_model='eval_loss',
            logging_steps=1,
        )
        trainer = transformers.Trainer(
            model=self.model,
            args=args,
            train_dataset=BLUWWMFSDataset(self.train_dataset),
            eval_dataset=BLUWWMFSDataset(self.val_dataset)
        )
        trainer.train()

    def evaluate(self):
        print('\nGETTING METRICS...\n')
        # Remove 'labels' key from validation dataset for prediction.
        outputs = self.model.to(self.val_dataset['input_ids'].device)(**self.val_dataset)

        predictions = self._decode_pos2strs(outputs['start_logits'].argmax(1), outputs['end_logits'].argmax(1), self.val_dataset['input_ids'])
        references = self._decode_pos2strs(self.val_dataset['start_positions'], self.val_dataset['end_positions'], self.val_dataset['input_ids'])

        metrics_funcs = [
            evaluate.load('exact_match'),
            evaluate.load('rouge')
        ]
        metrics = {}
        for metric in metrics_funcs:
            metrics.update(metric.compute(predictions=predictions, references=references))
        for k in metrics.keys():
            metrics[k] = metrics[k].item()
        return metrics

    def predict(self):
        print(f'\n({self.num_steps}/{self.num_steps}) CREATING TAILWIZ LABELS...\n')
        self.model = self.model.to(self.test_dataset['input_ids'].device)
        out_predictions = []
        for i in tqdm.tqdm(range(self.test_dataset['input_ids'].shape[0])):
            outputs = self.model(
                input_ids=self.test_dataset['input_ids'][i].unsqueeze(0),
                token_type_ids=self.test_dataset['token_type_ids'][i].unsqueeze(0),
                attention_mask=self.test_dataset['attention_mask'][i].unsqueeze(0))
            out_predictions.extend(self._decode_pos2strs(outputs['start_logits'].argmax(1), outputs['end_logits'].argmax(1), self.test_dataset['input_ids'][i].unsqueeze(0)))
        return out_predictions
        # return self._decode_pos2strs(outputs['start_logits'].argmax(1), outputs['end_logits'].argmax(1), self.test_dataset['input_ids'])


def parse(to_parse, labeled_examples=None, output_metrics=False):
    assert isinstance(to_parse, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
    assert 'prompt' in to_parse.columns and 'context' in to_parse.columns, \
        'Make sure the prompt column in your pandas DataFrame is named "prompt" and the context column is named "context".'

    if labeled_examples is not None:
        assert isinstance(labeled_examples, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
        assert 'prompt' in labeled_examples.columns and 'context' in labeled_examples.columns and 'label' in labeled_examples.columns, \
            'Make sure the prompt column in your pandas DataFrame is named "prompt", the context column is named "context", and the label column is named "label".'
    if output_metrics:
        assert labeled_examples is not None, 'In order to output an estimate of performance with output_metrics, labeled_examples must be provided.'

    if labeled_examples is None:
        parse_task_out = ParsingTask(None, None, to_parse)
        pred_results = parse_task_out.predict()
    else:
        assert len(labeled_examples) >= 2, 'At least 2 rows of prelabeled data must be given.'
        train, val = train_test_split(labeled_examples, test_size=0.2)
        parse_task_out = ParsingTask(train, val, to_parse)
        parse_task_out.train()
        pred_results = parse_task_out.predict()
    
    results = to_parse.copy()
    results['tailwiz_label'] = pred_results

    if output_metrics:
        metrics = parse_task_out.evaluate()

    print('\nDONE')
    return (results, metrics) if output_metrics else results



def list_tasks():
    return '''classify: Given text, classify the text.
    parse: Given a prompt and a context, parse the answer from the context.
    generate: Given a prompt, generate an answer.'''


import evaluate
import os
os.environ['TRANSFORMERS_CACHE'] = 'cache'  # Must be before transformers import.
import pandas as pd
import torch
import tqdm
import transformers
from sklearn.model_selection import train_test_split

from .task import Task

class GenerateTask(Task):
    def __init__(self, train, val, test):
        self.num_steps = 2 if train is None else 3
        print(f'\n(1/{self.num_steps}) PROCESSING DATA...\n')
        (
            self.train_dataset,
            self.val_dataset,
            self.val_references,
            self.test_dataset
        ) = self._load_data(train, val, test)
        self.model = self._load_model()

    def _load_data(self, train, val, test):
        self.tokenizer = transformers.T5Tokenizer.from_pretrained('google/flan-t5-base', cache_dir='cache/flan-t5')
        
        print('  - Data Processing Step 1 of 3...')
        train_tokens = None
        if train is not None:
            with tqdm.tqdm(total=2) as pbar:
                train_tokens = self.tokenizer(train.prompt.tolist(), return_tensors='pt', padding=True)
                pbar.update()
                train_labels = self.tokenizer(train.label.tolist(), return_tensors='pt', padding=True)
                train_tokens['labels'] = train_labels['input_ids']
                pbar.update()
        
        print('  - Data Processing Step 2 of 3...')
        val_tokens = None
        if val is not None:
            with tqdm.tqdm(total=2) as pbar:
                val_tokens = self.tokenizer(val.prompt.tolist(), return_tensors='pt', padding=True)
                pbar.update()
                val_labels = self.tokenizer(val.label.tolist(), return_tensors='pt', padding=True)
                val_tokens['labels'] = val_labels['input_ids']
                pbar.update()

        print('  - Data Processing Step 3 of 3...')
        with tqdm.tqdm(total=1) as pbar:
            test_tokens = self.tokenizer(test.prompt.tolist(), return_tensors='pt', padding=True)
            pbar.update()

        return train_tokens, val_tokens, (val.label.tolist() if val is not None else None), test_tokens

    def _load_model(self):
        return transformers.T5ForConditionalGeneration.from_pretrained('google/flan-t5-base', cache_dir='cache/flan-t5')

    def train(self):
        print(f'\n(2/{self.num_steps}) LEARNING...\n')
        class FT5Dataset(torch.utils.data.Dataset):
            def __init__(self, data):
                self.data = data
            
            def __len__(self):
                return len(self.data['input_ids'])
            
            def __getitem__(self, index):
                return {k: self.data[k][index] for k in self.data.keys()}

        args = transformers.Seq2SeqTrainingArguments(
            'cache/flan-t5',
            num_train_epochs=10,
            evaluation_strategy='epoch',
            save_strategy='epoch',
            load_best_model_at_end=True,
            predict_with_generate=True,
            metric_for_best_model='eval_loss',
            logging_steps=1,
        )
        trainer = transformers.Seq2SeqTrainer(
            model=self.model,
            args=args,
            train_dataset=FT5Dataset(self.train_dataset),
            eval_dataset=FT5Dataset(self.val_dataset) if self.val_dataset is not None else FT5Dataset(self.train_dataset),  # Requires eval dataset.
        )
        trainer.train()

    def evaluate(self):
        print('\nGETTING METRICS...\n')
        # Remove 'labels' key from validation dataset for prediction.
        dummy_val = self.val_dataset.copy()
        del dummy_val['labels']
        outputs = self.model.to(dummy_val['input_ids'].device).generate(**dummy_val)
        predictions = self.tokenizer.batch_decode(outputs, skip_special_tokens=True)

        metrics_funcs = [
            evaluate.load('exact_match'),
            evaluate.load('rouge')
        ]
        metrics = {}
        for metric in metrics_funcs:
            metrics.update(metric.compute(predictions=predictions, references=self.val_references))
        for k in metrics.keys():
            metrics[k] = metrics[k].item()
        return metrics

    def predict(self):
        print(f'\n({self.num_steps}/{self.num_steps}) CREATING TAILWIZ LABELS...\n')
        with tqdm.tqdm(total=1) as pbar:
            outputs = self.model.to(self.test_dataset['input_ids'].device).generate(**self.test_dataset)
            pbar.update()
        return self.tokenizer.batch_decode(outputs, skip_special_tokens=True)


def generate(to_generate, labeled_examples=None, output_metrics=False):
    assert isinstance(to_generate, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
    assert 'prompt' in to_generate.columns, \
        'Make sure the prompt column in your pandas DataFrame is named "prompt".'
    if labeled_examples is not None:
        assert isinstance(labeled_examples, pd.DataFrame), 'Make sure you are passing in pandas DataFrames.'
        assert 'prompt' in labeled_examples.columns and 'label' in labeled_examples.columns, \
            'Make sure the prompt column in your pandas DataFrame is named "prompt" and the label column is named "label".'
    if output_metrics:
        assert labeled_examples is not None, 'In order to output an estimate of performance with output_metrics, labeled_examples must be provided.'

    if labeled_examples is None:
        generate_task_out = GenerateTask(None, None, to_generate)
        pred_results = generate_task_out.predict()
    else:
        assert len(labeled_examples) >= 2, 'At least 2 rows of prelabeled data must be given.'
        train, val = train_test_split(labeled_examples, test_size=0.2)
        generate_task_out = GenerateTask(train, val, to_generate)
        generate_task_out.train()
        pred_results = generate_task_out.predict()
    
    results = to_generate.copy()
    results['tailwiz_label'] = pred_results

    if output_metrics:
        metrics = generate_task_out.evaluate()

    print('\nDONE')
    return (results, metrics) if output_metrics else results



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='tailwiz',
    version='0.0.13',
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'scikit-learn',
        'transformers',
        'evaluate',
        'pandas',
        'torch',
        'numpy',
        'sentencepiece',
    ],
    packages=find_packages(),
    author='Timothy Dai',
    author_email='timdai@stanford.edu',
    url='https://github.com/timothydai/tailwiz', 
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    package_dir={'': '.'},
)
